'''
# Simple Nets Core
Simplified module of net security concept for beginers.
'''
from .nets_core import (
    GrayHackApi, 
    WhiteHackApi, 
    DatsSecureApi, 
    HackModuleApi, 
    HashCryptoApi, 
    RandCryptoApi
)
